(* Title:  interpreter-state for Lucas-Interpretation
   Author: Walther Neuper 191017
   (c) due to copyright terms
*)
signature ENVIRONMENT =
sig
  type T
  val empty: T
  type T_TEST
  val T_TEST_to_string: T_TEST -> string

  val env2str: subst -> string
  val env2str': subst -> string
  val subst2str: subst -> string
  val subst2str': subst -> string
  val update: T -> term * term -> T
  val update': term -> term * T -> T
  val update_opt: T -> term option * term -> T
  val update_opt': T * (term option * term) -> T
  val update_opt'': ((term * T) * term option) -> T

end

(**)
structure Env(**): ENVIRONMENT(**) =
struct
(**)
type T = LibraryC.subst;
val empty = []: T
type T_TEST = (term * term list) list

fun T_TEST_to_string env_TEST =
  let
    val ctxt = ContextC.for_ERROR ()
  in
    env_TEST |> map (fn (t, ts) => "(" ^ UnparseC.term ctxt t ^ ", " ^ UnparseC.terms ctxt ts^ ")")
      |> strs2str'
  end

fun subst2str s =
  let
    val ctxt = Proof_Context.init_global (Know_Store.get_via_last_thy "Isac_Knowledge")
  in
    (strs2str o
      (map (
        linefeed o pair2str o
          (apsnd (UnparseC.term ctxt)) o (apfst (UnparseC.term ctxt))))) s
  end;
fun subst2str' s =
  let
    val ctxt = Proof_Context.init_global (Know_Store.get_via_last_thy "Isac_Knowledge")
  in
    (strs2str' o
     (map (
       pair2str o (apsnd (UnparseC.term ctxt)) o (apfst (UnparseC.term ctxt))))) s
  end;
val env2str = subst2str
val env2str' = subst2str'

(* update environment; t <> empty if coming from listexpr *)
fun update env (id, vl) =
  let val env' = if vl = TermC.empty then env else overwrite (env, (id, vl));
  in env' end;
fun update' id (vl, env) = update env (id, vl)

(*WN020526: not clear, when a is available in scan_up1 for eval_true*)
(*WN060906: in "fun check_leaf" eg. uses "SOME M__"(from some PREVIOUS
  curried Rewrite) for CURRENT value (which may be different from PREVIOUS);
  thus "NONE" must be set at the end of currying (ill designed anyway)*)
fun update_opt env (SOME id, vl) = update env (id, vl)
  | update_opt env (NONE, _) =
    ((** )tracing ("*** update_opt: (NONE," ^ UnparseC.term @{context} vl ^ ")");( **) env);
fun update_opt' (env, (id, vl)) = update_opt env (id, vl)
fun update_opt'' ((vl, env), id) = update_opt env (id, vl)


(**)end(**)
